from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import Any, Self

from attrs import define, field
from attrs.setters import convert
from .util import RandomIPv6Address


def _ip_address_or_name(value: str) -> IPv4Address | IPv6Address | str:
    try:
        return ip_address(value)
    except ValueError:
        return value


@define(on_setattr=convert)
class Endpoint:
    """Wireguard endpoint.

    Pedantic implementation to support ``wireguard-tools`` endpoint format.
    """
    host: IPv4Address | IPv6Address | str = field(converter=_ip_address_or_name)
    port: int = field(default=51820, converter=int)

    @classmethod
    def convert(cls, value: str | dict | Self) -> Self:
        if isinstance(value, str):
            return cls(*value.rsplit(":", 1))
        if isinstance(value, dict):
            return cls(**value)
        if isinstance(value, Endpoint):
            return value
        raise TypeError(f"Cannot parse endpoint from {value!r}")


@define(kw_only=True, on_setattr=convert)
class BaseNode:
    """A node in a Wireguard mesh network.

    This is a base class meant to be used for configuration from dictionaries.

    ``Mesh.nodes`` is a dictionary of Node objects, which are converted to ``MeshNode`` objects
    in ``Mesh.__attrs_post_init__()``.
    """
    ssh: str | dict
    """SSH connection string or arguments to the node.
    
    It is recommended to use the default SSH config file to specify connection details.
    
    This field is passed directly to ``fabric.Connection``, so any valid connection string
    can be used.
    """

    wg: Endpoint = field(converter=Endpoint.convert)
    """Public or mesh-accessible Wireguard endpoint."""

    port: int = field(default=51820, converter=int)
    """Wireguard listen port."""

    json: dict[str, Any] = field(factory=dict)
    """Set ``friendly_json`` data for all peers."""

    addr: IPv4Address | IPv6Address | RandomIPv6Address = field(
        converter=lambda x: ip_address(x) if not isinstance(x, (IPv4Address, IPv6Address)) else x,
        factory=RandomIPv6Address,
    )
    """Wireguard address of the node. If not specified, a random IPv6 address will be generated.
    
    This address is only used to establish a GRE tunnel between mesh nodes.
    It is not used for routing, and should not be changed once the mesh is created (configs written).
    """

    prio: int | None = field(default=None, validator=[lambda _, __, x: x is None or -8 <= x <= 7])
    """Bridge priority for GRE tunnels, normalized to ``-8 <= prio <= 7``.
    
    Calculated as ``32768 + 4096 * prio``.
    
    This field is used to create the initial bridge configurations, and should not be changed.
    
    When None, the bridge priority will be calculated as ``-8 + (index % 16)``.
    - Having any difference in bridge priority between nodes helps to avoid bad routing decisions.
    """



from functools import cached_property
from itertools import islice
from ipaddress import IPv4Interface, IPv6Interface, ip_interface
from secrets import token_bytes
import sys

from attrs import define
from attrs.setters import convert
from wireguard_tools import WireguardConfig, WireguardKey
from wireguard_tools.wireguard_config import WireguardPeer

from .types import BaseNode
from .remote import WireguardRemote
from .gretap import gretap_up, gretap_down
from .util import RandomIPv6Address


@define(kw_only=True, slots=False, on_setattr=convert)
class MeshNode(BaseNode):
    mesh: "mesh.Mesh"
    index: int
    config: WireguardConfig | None = None

    @cached_property
    def remote(self) -> WireguardRemote:
        return WireguardRemote(
            connect=self.ssh,
            interface=f"wg-{self.mesh.name}{self.index}",
        )

    @property
    def tag(self) -> str:
        return f"{self.network_addr.ip}@{self.wg.host}:{self.wg.port}"

    @property
    def network_addr(self) -> IPv4Interface | IPv6Interface:
        return ip_interface(f"{islice(self.mesh.network.hosts(), self.index, None).__next__()}/{self.mesh.network.prefixlen}")

    @property
    def bridge_priority(self) -> int:
        prio = self.prio if self.prio is not None else -8 + self.index % 16
        return 32768 + 4096 * prio

    @property
    def info(self) -> dict:
        return dict(
            host=self.remote.host,
            is_up=self.remote.is_up,
            config_exists=self.remote.config_exists,
            address=str(self.network_addr),
        )

    def __attrs_post_init__(self):
        self.__config_init()

    def __config_init(self):
        self.config = self.remote.config

        if self.config is not None:
            if isinstance(self.addr, RandomIPv6Address):
                self.addr = self.config.addresses[0].ip
            if self.config.listen_port is not None:
                self.port = self.config.listen_port

        else:
            self.config = WireguardConfig(
                private_key=WireguardKey.generate(),
                addresses=[ip_interface(self.addr)],
                listen_port=self.port,
            )

    @classmethod
    def from_node(cls, mesh: "mesh.Mesh", index: int, node: BaseNode) -> "MeshNode":
        return cls(
            mesh=mesh,
            index=index,
            ssh=node.ssh,
            wg=node.wg,
            port=node.port,
            addr=node.addr,
            json=node.json,
        )

    def can_peer(self, other: "MeshNode") -> bool:
        """Check if this node can peer with another node.

        Uses the Wireguard UDP listen ports and endpoints to check if the nodes can peer,
        so this can only be reliable when Wireguard is not up on either node.
        """
        return (not self.remote.is_up and self.remote.udping_from(
            listen_port=self.port,
            endpoint_host=self.wg.host,
            endpoint_port=self.wg.port,
            remote=other.remote,
        )) or (not other.remote.is_up and other.remote.udping_from(
            listen_port=other.port,
            endpoint_host=other.wg.host,
            endpoint_port=other.wg.port,
            remote=self.remote,
        ))

    def as_peer(self, **kwds) -> WireguardPeer:
        return WireguardPeer(
            public_key=self.config.private_key.public_key(),
            allowed_ips=[ip_interface(self.addr)],
            endpoint_host=self.wg.host,
            endpoint_port=self.wg.port,
            friendly_name=self.remote.host,
            friendly_json=self.json or None,
            **kwds,
        )

    def peer_with(self, other: "MeshNode"):
        self_pubkey = self.config.private_key.public_key()
        other_pubkey = other.config.private_key.public_key()

        if self_pubkey == other_pubkey:
            raise ValueError("Cannot peer with self")

        if self_pubkey in other.config.peers and other_pubkey in self.config.peers:
            self_peer = other.config.peers[self_pubkey]
            other_peer = self.config.peers[other_pubkey]

            if self_peer.allowed_ips[0].ip != self.addr or other_peer.allowed_ips[0].ip != other.addr:
                raise ValueError("Existing peering has different addresses")

            # noinspection DuplicatedCode
            if not self.json and self_peer.friendly_json is not None:
                self.json = self_peer.friendly_json
            elif self.json and self_peer.friendly_json != self.json:
                self_peer.friendly_json = self.json

            if self_peer.friendly_name != self.remote.host:
                self_peer.friendly_name = self.remote.host

            # noinspection DuplicatedCode
            if not other.json and other_peer.friendly_json is not None:
                other.json = other_peer.friendly_json
            elif other.json and other_peer.friendly_json != other.json:
                other_peer.friendly_json = other.json

            if other_peer.friendly_name != other.remote.host:
                other_peer.friendly_name = other.remote.host

            return

        if not self.mesh.full and not self.can_peer(other):
            return

        preshared_key = WireguardKey(token_bytes(32))
        self_peer = self.as_peer(preshared_key=preshared_key)
        other_peer = other.as_peer(preshared_key=preshared_key)

        self.config.add_peer(other_peer)
        other.config.add_peer(self_peer)

        self_index = self.index
        other_index = other.index
        self_gretap_name = f"{self.mesh.name}{self_index}{other_index}"
        other_gretap_name = f"{other.mesh.name}{other_index}{self_index}"
        self_gretap_addr = self.addr
        other_gretap_addr = other.addr
        self_bridge_name = f"{self.mesh.name}{self_index}"
        other_bridge_name = f"{other.mesh.name}{other_index}"
        self_bridge_addr = self.network_addr
        other_bridge_addr = other.network_addr

        self.config.postup.extend(gretap_up(
            gretap_name=self_gretap_name,
            bridge_name=self_bridge_name,
            priority=self.bridge_priority,
            local=self_gretap_addr,
            remote=other_gretap_addr,
            bridge_addr=self_bridge_addr,
        ))

        other.config.postup.extend(gretap_up(
            gretap_name=other_gretap_name,
            bridge_name=other_bridge_name,
            priority=other.bridge_priority,
            local=other_gretap_addr,
            remote=self_gretap_addr,
            bridge_addr=other_bridge_addr,
        ))

        self.config.predown.extend(gretap_down(
            gretap_name=self_gretap_name,
            bridge_name=self_bridge_name,
        ))

        other.config.predown.extend(gretap_down(
            gretap_name=other_gretap_name,
            bridge_name=other_bridge_name,
        ))

    def up(self, *, write: bool | None = None) -> bool:
        remote: WireguardRemote = self.remote

        if remote.is_up:
            remote.down()

        if write or write is None and (write := not remote.config_exists) or (write := remote.config != self.config):
            try:
                remote.config_write(self.config)

            except Exception as exc:
                print(f"[{self.tag}] [up] !! config_write failed: {exc}", file=sys.stderr)
                return False

        if isinstance(up := remote.up(), RuntimeError):
            print(f"[{self.tag}] [up] !! {remote.interface}:\n{up}", file=sys.stderr)
            if write:
                remote.config_remove()
            return False

        else:
            print(f"[{self.tag}] [up] ++ {remote.interface}:\n{up}", file=sys.stderr)

        return True

    def down(self, *, remove: bool | None = None) -> bool:
        remote: WireguardRemote = self.remote

        if remote.is_up:
            if isinstance(down := remote.down(), RuntimeError):
                print(f"[{self.tag}] [down] !! {remote.interface}:\n{down}", file=sys.stderr)

                if not remote.is_up and remove:
                    remote.config_remove()
                return False

            else:
                print(f"[{self.tag}] [down] -- {remote.interface}:\n{down}", file=sys.stderr)

        if remove or remove is None and remote.config_exists:
            remote.config_remove()

        return True

    def sync(self, *, up: bool | None = None) -> bool:
        remote: WireguardRemote = self.remote
        remote_config = remote.config

        if remote_config is None or remote_config != self.config:
            if up or up is None and remote.is_up:
                return self.up(write=True)

            remote.config_write(self.config)
            return True

        return False



from itertools import combinations
from ipaddress import IPv6Address, IPv4Network, IPv6Network, ip_network

from attrs import define, field
from attrs.setters import convert

from .types import BaseNode
from .node import MeshNode


@define(kw_only=True, on_setattr=convert)
class Mesh:
    name: str

    nodes: dict[int, MeshNode] = field(
        converter=lambda nodes_: {
            int(index): BaseNode(**node) if not isinstance(node, BaseNode) else node
            for index, node in nodes_.items()
        },
    )

    network: IPv4Network | IPv6Address = field(
        converter=lambda x: ip_network(x) if not isinstance(x, (IPv4Network, IPv6Network)) else x,
    )
    """Mesh network address.
    
    This network will be routed among all node bridges in the mesh.
    """

    full: bool = True
    """Peer all node pairs regardless of reachability.
    
    When False, checks whether two nodes can reach each other before connecting them.
    """

    @property
    def pairs(self):
        return combinations(self.nodes.values(), 2)

    @property
    def is_up(self) -> float:
        return sum(int(node.remote.is_up) for node in self.nodes.values()) / len(self.nodes)

    @property
    def config_exists(self) -> float:
        return sum(int(node.remote.config_exists) for node in self.nodes.values()) / len(self.nodes)

    @property
    def info(self) -> dict:
        return dict(
            name=self.name,
            network=str(self.network),
            is_up=self.is_up,
            config_exists=self.config_exists,
            nodes={index: node.info for index, node in self.nodes.items()},
        )

    def __attrs_post_init__(self):
        for index, node in self.nodes.items():
            self.nodes[index] = MeshNode.from_node(self, index, node) if not isinstance(node, MeshNode) else node

        for node1, node2 in self.pairs:
            node1.peer_with(node2)

    def __iter__(self):
        return iter(self.nodes)

    def __getitem__(self, item):
        return self.nodes[item]

    def __len__(self):
        return len(self.nodes)

    def config_write(self) -> None:
        for node in self.nodes.values():
            node.remote.config_write(node.config)

    def config_remove(self) -> None:
        for node in self.nodes.values():
            node.remote.config_remove()

    def up(self, *, write: bool | None = None) -> bool | None:
        up_nodes = []

        for node in self.nodes.values():
            if node.up(write=write):
                up_nodes.append(node)
            else:
                for up_node in up_nodes:
                    up_node.down(remove=write)
                return False

        return None if not up_nodes else True

    def down(self, *, remove: bool | None = None) -> bool:
        return all(node.down(remove=remove) for node in self.nodes.values())

    def sync(self, *, up: bool | None = None) -> bool:
        return all(node.sync(up=up) for node in self.nodes.values())

    def show(self) -> None:
        for node in self.nodes.values():
            print(f"[{node.tag}]\n{node.remote.show()}\n")
